#define DBG 0
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <signal.h>

#include "host.h"
#include "misc.h"
#include "machine.h"
#include "regs.h"
#include "memory.h"
#include "cache.h"
#include "loader.h"
#include "syscall.h"
#include "bpred.h"
#include "resource.h"
#include "bitmap.h"
#include "options.h"
#include "eval.h"
#include "stats.h"
#include "ptrace.h"
#include "dlite.h"
#include "sim.h"

#include "bloom.h"
#include "region.h"

region_t *
region_create (char *args, struct stat_sdb_t *sdb)
{
  char *h, *s;
  int done = 0;
  int i = 0;
  region_t *region;
  int rsize, bsize, sets, ass, vsize;
  char ts[256];
  char vn[256];
  char fn[256];

  if (strcmp (args, "none") == 0) return 0;
  region = (region_t *) calloc (1, sizeof (region_t));
  assert (region);

  fprintf (stderr, "region_create: args (%s) ", args);

  //--------------------------------------------------------------------------------
  // args = NAME:CRH_CBV
  // NAME = string
  // CRH = MBLOOM = START1-SIZE1,START2-SIZE2,...,STARTn-SIZEn
  // CBV = REGIONSIZE,BLOCKSIZE,SETS,ASSOCIATIVITY
  //--------------------------------------------------------------------------------
  // get name
  //--------------------------------------------------------------------------------
  s = args;
  while (*s != ':') s++;
  *s = 0;
  region->name = (char *) calloc (strlen (args) + 2,  sizeof (char));
  strcpy (region->name+1, args);
  *region->name = 'R';
  fprintf (stderr, " name %s\n", region->name);
  //--------------------------------------------------------------------------------
  // create CRH
  //--------------------------------------------------------------------------------
  *s = ':';
  s++;
  done = 0;
  while (*s != '_') s++;
  *s = 0;
  region->crh = mbloom_create (args, sdb);
  s++;
  //--------------------------------------------------------------------------------
  // create CBV
  //--------------------------------------------------------------------------------
  sscanf (s, "%d,%d,%d,%d", &rsize, &bsize, &sets, &ass);
  sprintf (ts, "%s_CBV", region->name);
  vsize = (rsize/bsize) / 8;
  region->cbv = cache_create (ts, sets, rsize, 0, vsize, ass, LRU, NULL, 1); 
  fprintf (stderr, "\tcbv: rsize %d bsize %d sets %d assoc %d vsize %d (bits)\n", rsize, bsize, sets, ass, vsize*8);
  region->rsize = rsize;
  region->bsize = bsize;
  region->bshift = 0;
  bsize >>= 1;
  while (bsize)
  {
    region->bshift++;
    bsize >>= 1;
  }

   sprintf (vn, "%s_region_probes", region->name); stat_reg_counter(sdb, vn, "probes", &region->probes, region->probes, NULL);
   sprintf (vn, "%s_region_updates", region->name); stat_reg_counter(sdb, vn, "updates", &region->updates, region->updates, NULL);
   sprintf (vn, "%s_outcome000", region->name); stat_reg_counter(sdb, vn, "real/crh/cbv", &region->outcome[0][0][0], region->outcome[0][0][0], NULL);
   sprintf (vn, "%s_outcome001", region->name); stat_reg_counter(sdb, vn, "real/crh/cbv", &region->outcome[0][0][1], region->outcome[0][0][1], NULL);
   sprintf (vn, "%s_outcome010", region->name); stat_reg_counter(sdb, vn, "real/crh/cbv", &region->outcome[0][1][0], region->outcome[0][1][0], NULL);
   sprintf (vn, "%s_outcome011", region->name); stat_reg_counter(sdb, vn, "real/crh/cbv", &region->outcome[0][1][1], region->outcome[0][1][1], NULL);
   sprintf (vn, "%s_outcome100", region->name); stat_reg_counter(sdb, vn, "real/crh/cbv", &region->outcome[1][0][0], region->outcome[1][0][0], NULL);
   sprintf (vn, "%s_outcome101", region->name); stat_reg_counter(sdb, vn, "real/crh/cbv", &region->outcome[1][0][1], region->outcome[1][0][1], NULL);
   sprintf (vn, "%s_outcome110", region->name); stat_reg_counter(sdb, vn, "real/crh/cbv", &region->outcome[1][1][0], region->outcome[1][1][0], NULL);
   sprintf (vn, "%s_outcome111", region->name); stat_reg_counter(sdb, vn, "real/crh/cbv", &region->outcome[1][1][1], region->outcome[1][1][1], NULL);

   sprintf (vn, "%s_misses", region->name); sprintf (fn, "%s_outcome000 + %s_outcome010", region->name, region->name);
   stat_reg_formula(sdb, vn, "percentage of all accesses predicted as misses", fn, NULL);
   sprintf (vn, "%s_real_misses", region->name); sprintf (fn, "%s_misses + %s_outcome001 + %s_outcome011", region->name, region->name, region->name);
   stat_reg_formula(sdb, vn, "percentage of all accesses predicted as misses", fn, NULL);
   sprintf (vn, "%s_real_hits", region->name); sprintf (fn, "%s_outcome100 + %s_outcome101 + %s_outcome110 + %s_outcome111", region->name, region->name, region->name, region->name);
   stat_reg_formula(sdb, vn, "percentage of all accesses predicted as misses", fn, NULL);

   sprintf (vn, "%s_coverage", region->name); sprintf (fn, "%s_misses / %s_real_misses", region->name, region->name);
   stat_reg_formula(sdb, vn, "percentage of misses predicted", fn, NULL);
   sprintf (vn, "%s_overall", region->name); sprintf (fn, "%s_misses / (%s_real_misses+ %s_real_hits)", region->name, region->name, region->name);
   stat_reg_formula(sdb, vn, "percentage of all accesses predicted as misses", fn, NULL);

   return region;
}

int
region_array_create (region_t *array[], char *args, struct stat_sdb_t *sdb)
{
  char *h;
  int done = 0;
  int i = 0;
  if (strcmp (args, "none") == 0) return 0;

  h = args;

  while (!done && args)
  {
	h = args;
	while (*args != '#' && *args) args++;
	if (*args == '#') *args = 0;
	else done = 1;
	args++;
	array[i++] = region_create (h, sdb);
  }
  return i;
}


static inline unsigned int
region_blk_offset (region_t *rp, md_addr_t addr)
{
   unsigned long idx = (addr & (rp->rsize - 1)) >> rp->bshift;

	if (DBG) fprintf (stderr, "region_blk_offset: 0x%08lx idx 0x%08lx\n", addr, idx);
   return idx;
}

static inline unsigned int
region_vector_idx (region_t *rp, md_addr_t addr)
{
	unsigned long idx = region_blk_offset (rp, addr) >> 3; 
	if (DBG) fprintf (stderr, "region_vector_idx: 0x%08lx idx 0x%08lx\n", addr, idx);
	return idx;
}

static inline unsigned int
region_vector_mask (region_t *rp, md_addr_t addr)
{
	unsigned long idx =  (1 << (region_blk_offset (rp, addr) & 7)); 
	if (DBG) fprintf (stderr, "region_vector_mask: 0x%08lx idx 0x%08lx\n", addr, idx);
	return idx;
}

int
cbv_probe (region_t *rp, md_addr_t addr)
{
	byte_t *vector;

	if (!cache_probe (rp->cbv, addr)) return -1;
	cache_access (rp->cbv, Read, addr, NULL, 4, 0, &vector, NULL);
	return (vector[region_vector_idx (rp, addr)] & region_vector_mask (rp, addr)) ? 1 : 0;
}

void
cbv_block_dealloc (region_t *rp, md_addr_t addr)
{
	byte_t *vector;

	if (!cache_probe (rp->cbv, addr)) return;
	cache_access (rp->cbv, Read, addr, NULL, 4, 0, &vector, NULL);

	vector[region_vector_idx (rp, addr)] &= ~region_vector_mask (rp, addr);
}

void
cbv_block_alloc (region_t *rp, md_addr_t addr)
{
	byte_t *vector;

	if (!cache_probe (rp->cbv, addr)) return;
	cache_access (rp->cbv, Read, addr, NULL, 4, 0, &vector, NULL);

	vector[region_vector_idx (rp, addr)] |= region_vector_mask (rp, addr);
}

void
cbv_region_alloc (region_t *rp, md_addr_t addr)
{
	byte_t *vector;
 	int i;

	if (cache_probe (rp->cbv, addr)) return;
	cache_access (rp->cbv, Read, addr, NULL, 4, 0, &vector, NULL);

 	for (i = 0; i < rp->rsize/rp->bsize/8; i++) vector[i] = 0;

	vector[region_vector_idx (rp, addr)] |= region_vector_mask (rp, addr);
}

void
region_alloc (region_t *rp, md_addr_t addr)
{
	int i, crh_hit, cbv_hit;
	int rv;

	rp->updates++;
	crh_hit = mbloom_probe (rp->crh, addr, 0);
	cbv_hit = cbv_probe (rp, addr);

	if (!crh_hit  && cbv_hit == -1) 
	{
		cbv_region_alloc (rp, addr);
	}
	cbv_block_alloc (rp, addr);
	mbloom_alloc (rp->crh, addr);
}

void
region_dealloc (region_t *rp, md_addr_t addr)
{
	int i, crh_hit, cbv_hit;
        int rv;

        rp->updates++;
	mbloom_dealloc (rp->crh, addr);
        cbv_block_dealloc (rp, addr);
}


int
region_probe (region_t *rp, md_addr_t addr, int real_hit)
{
	int i, crh_hit, cbv_hit;
	int rv;

	rp->probes++;
	crh_hit = mbloom_probe (rp->crh, addr, 0);
	cbv_hit = cbv_probe (rp, addr);
	assert (crh_hit == 0 || crh_hit == 1);
	assert (cbv_hit >= -1 || cbv_hit <= 1);

	if (!crh_hit  && cbv_hit == -1) rv = 0;	// CRH says miss and no entry in CBV
	else rv = cbv_hit;
	rp->outcome[real_hit][crh_hit][(cbv_hit != 0) ? 1 : 0]++;


	if (real_hit && !rv)
		fatal ("region_probe: real hit ->miss ? addr 0x%08lx, crh %d, cbv %d\n", addr, crh_hit, cbv_hit);

	return rv;
}

